@@ -4,13 +4,27 @@ import os.path as op
 
 import click
 
-from .base import lgr, map_to_click_exceptions
+from .base import devel_option, lgr, map_to_click_exceptions
+from ..dandiarchive import DandisetURL, _dandi_url_parser, parse_dandi_url
+from ..misctypes import Digest
 from ..utils import is_url
 
 # TODO: all the recursion options etc
 
 
-@click.command()
+# The use of f-strings apparently makes this not a proper docstring, and so
+# click doesn't use it unless we explicitly assign it to `help`:
+@click.command(
+    help=f"""\
+List .nwb files and dandisets metadata.
+
+The arguments may be either resource identifiers or paths to local
+files/directories.
+
+\b
+{_dandi_url_parser.known_patterns}
+"""
+)
 @click.option(
     "-F",
     "--fields",
@@ -23,7 +37,7 @@ from ..utils import is_url
     "--format",
     help="Choose the format/frontend for output. If 'auto', 'pyout' will be "
     "used in case of multiple files, and 'yaml' for a single file.",
-    type=click.Choice(["auto", "pyout", "json", "json_pp", "yaml"]),
+    type=click.Choice(["auto", "pyout", "json", "json_pp", "json_lines", "yaml"]),
     default="auto",
 )
 @click.option(
@@ -50,14 +64,35 @@ from ..utils import is_url
     help="Convert metadata to new schema version",
     metavar="VERSION",
 )
-@click.argument("paths", nargs=-1, type=click.Path(exists=False, dir_okay=True))
+@devel_option(
+    "--use-fake-digest",
+    is_flag=True,
+    help="Use dummy value for digests of local files instead of computing",
+)
+@click.argument(
+    "paths", nargs=-1, type=click.Path(exists=False, dir_okay=True), metavar="PATH|URL"
+)
 @map_to_click_exceptions
-def ls(paths, schema, metadata, fields=None, format="auto", recursive=False, jobs=6):
+def ls(
+    paths,
+    schema,
+    metadata,
+    use_fake_digest=False,
+    fields=None,
+    format="auto",
+    recursive=False,
+    jobs=6,
+):
     """List .nwb files and dandisets metadata."""
+
     # TODO: more logical ordering in case of fields = None
-    from .formatter import JSONFormatter, PYOUTFormatter, YAMLFormatter
+    from .formatter import (
+        JSONFormatter,
+        JSONLinesFormatter,
+        PYOUTFormatter,
+        YAMLFormatter,
+    )
     from ..consts import metadata_all_fields
-    from ..dandiapi import DandiAPIClient
 
     # TODO: avoid
     from ..support.pyout import PYOUT_SHORT_NAMES_rev
@@ -65,13 +100,13 @@ def ls(paths, schema, metadata, fields=None, format="auto", recursive=False, job
 
     common_fields = ("path", "size")
     if schema is not None:
-        from ..models import AssetMeta, DandisetMeta
+        from dandischema import models
 
         all_fields = tuple(
             sorted(
                 set(common_fields)
-                | DandisetMeta.__fields__.keys()
-                | AssetMeta.__fields__.keys()
+                | models.Dandiset.__fields__.keys()
+                | models.Asset.__fields__.keys()
             )
         )
     else:
@@ -100,37 +135,21 @@ def ls(paths, schema, metadata, fields=None, format="auto", recursive=False, job
     def assets_gen():
         for path in paths:
             if is_url(path):
-                from ..dandiarchive import navigate_url
-
-                with navigate_url(path) as (client, dandiset, assets):
-                    if dandiset:
-                        dandiset_id = dandiset.get("dandiset", {}).get("identifier")
+                parsed_url = parse_dandi_url(path)
+                with parsed_url.navigate() as (client, dandiset, assets):
+                    if isinstance(parsed_url, DandisetURL):
                         rec = {
-                            "path": dandiset.pop("dandiset", {}).get(
-                                "identifier", "ERR#%s" % id(dandiset)
-                            )
+                            "path": dandiset.identifier,
+                            **dandiset.version.json_dict(),
+                            "metadata": dandiset.get_raw_metadata(),
                         }
-                        # flatten the metadata into record to display
-                        # rec.update(dandiset.get('metadata', {}))
-                        rec.update(dandiset)
                         yield rec
-                    else:
-                        dandiset_id = None
-                    if recursive and assets:
-                        if isinstance(client, DandiAPIClient) and metadata in (
-                            "all",
-                            "assets",
-                        ):
-                            for a in assets:
-                                if "metadata" not in a:
-                                    a["metadata"] = client.get_asset(
-                                        dandiset_id,
-                                        dandiset["version"],
-                                        a["asset_id"],
-                                    )
-                                yield a
-                        else:
-                            yield from assets
+                    if not isinstance(parsed_url, DandisetURL) or recursive:
+                        for a in assets:
+                            rec = a.json_dict()
+                            if metadata in ("all", "assets"):
+                                rec["metadata"] = a.get_raw_metadata()
+                            yield rec
             else:
                 # For now we support only individual files
                 yield path
@@ -149,6 +168,8 @@ def ls(paths, schema, metadata, fields=None, format="auto", recursive=False, job
         out = JSONFormatter()
     elif format == "json_pp":
         out = JSONFormatter(indent=2)
+    elif format == "json_lines":
+        out = JSONLinesFormatter()
     elif format == "yaml":
         out = YAMLFormatter()
     else:
@@ -177,6 +198,7 @@ def ls(paths, schema, metadata, fields=None, format="auto", recursive=False, job
                             errors=errors,
                             flatten=format == "pyout",
                             schema=schema,
+                            use_fake_digest=use_fake_digest,
                         )
                         if format == "pyout":
                             rec[async_keys] = cb
@@ -305,45 +327,54 @@ def flatten_meta_to_pyout(meta):
     return out
 
 
-def get_metadata_ls(path, keys, errors, flatten=False, schema=None):
+def get_metadata_ls(
+    path, keys, errors, flatten=False, schema=None, use_fake_digest=False
+):
     from ..dandiset import APIDandiset
     from ..metadata import get_metadata, nwb2asset
     from ..pynwb_utils import get_nwb_version, ignore_benign_pynwb_warnings
+    from ..support.digests import get_digest
 
     ignore_benign_pynwb_warnings()
 
     def fn():
-        try:
-            rec = {}
-            # No need for calling get_metadata if no keys are needed from it
-            if keys is None or list(keys) != ["nwb_version"]:
-                try:
-                    if schema is not None:
-                        if op.isdir(path):
-                            dandiset = APIDandiset(path, schema_version=schema)
-                            rec = dandiset.metadata
-                        else:
-                            rec = nwb2asset(path, schema_version=schema).json_dict()
+        rec = {}
+        # No need for calling get_metadata if no keys are needed from it
+        if keys is None or list(keys) != ["nwb_version"]:
+            try:
+                if schema is not None:
+                    if op.isdir(path):
+                        dandiset = APIDandiset(path, schema_version=schema)
+                        rec = dandiset.metadata
                     else:
-                        rec = get_metadata(path)
-                except Exception as exc:
-                    _add_exc_error(path, rec, errors, exc)
-                if flatten:
-                    rec = flatten_meta_to_pyout(rec)
-            if keys is not None:
-                rec = {k: v for k, v in rec.items() if k in keys}
-            if (
-                not op.isdir(path)
-                and "nwb_version" not in rec
-                and (keys and "nwb_version" in keys)
-            ):
-                # Let's at least get that one
-                try:
-                    rec["nwb_version"] = get_nwb_version(path)
-                except Exception as exc:
-                    _add_exc_error(path, rec, errors, exc)
-            return rec
-        finally:
-            pass
+                        if use_fake_digest:
+                            digest = "0" * 32 + "-1"
+                        else:
+                            lgr.info("Calculating digest for %s", path)
+                            digest = get_digest(path, digest="dandi-etag")
+                        rec = nwb2asset(
+                            path,
+                            schema_version=schema,
+                            digest=Digest.dandi_etag(digest),
+                        ).json_dict()
+                else:
+                    rec = get_metadata(path)
+            except Exception as exc:
+                _add_exc_error(path, rec, errors, exc)
+            if flatten:
+                rec = flatten_meta_to_pyout(rec)
+        if keys is not None:
+            rec = {k: v for k, v in rec.items() if k in keys}
+        if (
+            not op.isdir(path)
+            and "nwb_version" not in rec
+            and (keys and "nwb_version" in keys)
+        ):
+            # Let's at least get that one
+            try:
+                rec["nwb_version"] = get_nwb_version(path)
+            except Exception as exc:
+                _add_exc_error(path, rec, errors, exc)
+        return rec
 
     return fn
